#include "map.hh"
#include "bsp.hh"

Map::Map() : Walls(NULL), WallCount(0)
{
}

Map::~Map()
{
    if(Walls)delete[] Walls;
}

#include <cstdio>
void Map::Load(const v3 *verts, unsigned vcount,
               const wall *walls, unsigned wcount)
{
    if(Walls) { delete[] Walls; Walls=NULL; }
    Walls = new MapWall[WallCount = wcount];
    FLOAT mix=0,miy=0,miz=0, max=0,may=0,maz=0;
    for(unsigned w=0; w<wcount; ++w)
    {
    	if(walls[w].v1 >= vcount || walls[w].v2 >= vcount)
    	{
    		std::fprintf(stderr, "Wall %u has vertex out of range (max %u, has %u and %u)\n",
    		    w, vcount, walls[w].v1, walls[w].v2);
    		continue;
    	}
        Walls[w] = MapWall(walls[w], verts);
        if(!w || Walls[w].minx()<mix)mix=Walls[w].minx();
        if(!w || Walls[w].miny()<miy)miy=Walls[w].miny();
        if(!w || Walls[w].minz()<miz)miz=Walls[w].minz();
        if(!w || Walls[w].maxx()>max)max=Walls[w].maxx();
        if(!w || Walls[w].maxy()>may)may=Walls[w].maxy();
        if(!w || Walls[w].maxz()>maz)maz=Walls[w].maxz();
        
        unsigned ccount=0;
        if(Walls[w].p1.x == Walls[w].p2.x)ccount++;
        if(Walls[w].p1.y == Walls[w].p2.y)ccount++;
        if(Walls[w].p1.z == Walls[w].p2.z)ccount++;
        if(ccount != 1)
        {
        	std::fprintf(stderr, "Wrong common count at wall %u (is %u, should be 1)\n", w, ccount);
        }
    }
    
    tree = new bsp(XYZ(mix,miy,miz), XYZ(max,may,maz), wcount);
    tree->Split(Walls);
}

MapWall::MapWall(const wall &w, const v3 *verts) : Pigm(w.R/255.0, w.G/255.0, w.B/255.0)
{
    const v3 &a = verts[w.v1], &b = verts[w.v2];
    p1.Set(a.x, a.y, a.z);
    p2.Set(b.x, b.y, b.z);
    
    Pigm.setdefault();
    Pigm.diffmul = w.diffper/100.0;
    Pigm.specmul = w.specper/100.0;
}

bool MapWall::hits(const XYZ &a, const XYZ &b) const
{
    return (a.x <= maxx() && b.x >= minx())
        && (a.y <= maxy() && b.y >= miny())
        && (a.z <= maxz() && b.z >= minz());
}

Map *LoadMap(const v3 *verts, unsigned vcount,
             const wall *walls, unsigned wcount)
{
    Map *tmp = new Map();
    tmp->Load(verts, vcount, walls, wcount);
    return tmp;
}

MapWall *Map::FindObstacle(const XYZ &eye, const XYZ &dir)
{
	return FindObstacleWith(eye, dir, tree);
}
MapWall *Map::FindObstacleWith(const XYZ &eye, const XYZ &dir, const bsp *tre)
{
	if(eye.x > tre->p2.x) { if(dir.x >= 0)return NULL; } else if(eye.x < tre->p1.x && dir.x <= 0)return NULL;
	if(eye.y > tre->p2.y) { if(dir.y >= 0)return NULL; } else if(eye.y < tre->p1.y && dir.y <= 0)return NULL;
	if(eye.z > tre->p2.z) { if(dir.z >= 0)return NULL; } else if(eye.z < tre->p1.z && dir.z <= 0)return NULL;
	
	bool omite = (eye.x <= tre->middle.x && dir.x < 0);
	bool omitw = (eye.x >= tre->middle.x && dir.x > 0);
	bool omitn = (eye.y <= tre->middle.y && dir.y < 0);
	bool omits = (eye.y >= tre->middle.y && dir.y > 0);
	bool omitu = (eye.z <= tre->middle.z && dir.z < 0);
	bool omitd = (eye.z >= tre->middle.z && dir.z > 0);
	
	for(unsigned a=0; a<8; ++a)
	{
		if((a&1) ? omits : omitn)continue;
		if((a&2) ? omite : omitw)continue;
		if((a&4) ? omitd : omitu)continue;
		MapWall *m = FindObstacleWith(eye, dir, tre->dirs[a]);
		if(m)return m;
	}
	for(unsigned a=0; a<tre->wallids.size(); ++a)
	{
		MapWall *m = &Walls[tre->wallids[a]];
		if(m->IsObstacle(eye, dir))
		{
			return m;
		}
	}
	return NULL;
}

bool MapWall::IsObstacle(const XYZ &eye, const XYZ &dir) const
{
	if(p1.x == p2.x)
	{
		FLOAT x = p1.x;
		/* Horizontal wall */
		
		if((eye.x < x && dir.x <= 0) || (eye.x > x && dir.x >= 0)) return false;
		
		FLOAT ypos = eye.y + (x-eye.x) * dir.y;
		FLOAT zpos = eye.z + (x-eye.x) * dir.z;
		if(ypos < miny() || ypos > maxy())return false;
		if(zpos < minz() || zpos > maxz())return false;
		
		/* Hit at x, ypos, zpos */
		return true;
	}
	if(p1.y == p2.y)
	{
		FLOAT y = p1.y;
		/* Vertical wall */
		
		if((eye.y < y && dir.y <= 0) || (eye.y > y && dir.y >= 0)) return false;
		
		FLOAT xpos = eye.x + (y-eye.y) * dir.x;
		FLOAT zpos = eye.z + (y-eye.y) * dir.z;
		if(xpos < minx() || xpos > maxx())return false;
		if(zpos < minz() || zpos > maxz())return false;
		
		/* Hit at xpos, y, zpos */
		return true;
	}
	if(p1.z == p2.z)
	{
		FLOAT z = p1.z;
		/* Horizontal wall */
		
		if((eye.z < z && dir.z <= 0) || (eye.z > z && dir.z >= 0)) return false;
		
		FLOAT xpos = eye.x + (z-eye.z) * dir.x;
		FLOAT ypos = eye.y + (z-eye.z) * dir.y;
		if(xpos < minx() || xpos > maxx())return false;
		if(ypos < miny() || ypos > maxy())return false;
		
		/* Hit at xpos, ypos, z */
		return true;
	}
	/* This isn't a wall, but a box! */
	return false;
}

const XYZ MapWall::HitPoint(const XYZ &eye, const XYZ &dir) const
{
	if(p1.x == p2.x)
	{
		FLOAT x = p1.x;
		/* Horizontal wall */
		
		FLOAT ypos = eye.y + (x-eye.x) * dir.y;
		FLOAT zpos = eye.z + (x-eye.x) * dir.z;
		return XYZ(x, ypos, zpos);
	}
	if(p1.y == p2.y)
	{
		FLOAT y = p1.y;
		/* Vertical wall */
		
		FLOAT xpos = eye.x + (y-eye.y) * dir.x;
		FLOAT zpos = eye.z + (y-eye.y) * dir.z;
		return XYZ(xpos, y, zpos);
	}
	if(p1.z == p2.z)
	{
		FLOAT z = p1.z;
		/* Horizontal wall */
		
		FLOAT xpos = eye.x + (z-eye.z) * dir.x;
		FLOAT ypos = eye.y + (z-eye.z) * dir.y;
		return XYZ(xpos, ypos, z);
	}
	return XYZ();
}
